# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
from collections import OrderedDict

from netman.core.objects.bond import Bond
from netman.core.objects.interface import Interface
from netman.core.objects.interface_states import OFF, ON
from netman.core.objects.port_modes import ACCESS, TRUNK
from netman.core.objects.switch_base import SwitchBase
from netman.core.objects.vlan import Vlan
from netman.core.objects.vrrp_group import VrrpGroup

__all__ = ['CachedSwitch']


class Cache(object):
    object_type = None
    object_key = None

    def __init__(self, *key_value_tuples):
        self.refresh_items = set()
        self.dict = OrderedDict(*key_value_tuples)

    def create_fake_object(self, item):
        params = {self.object_key: item}
        return self.object_type(**params)

    def invalidated(self):
        self.refresh_items.add(None)
        return self

    def __getitem__(self, item):
        try:
            return self.dict[item]
        except KeyError:
            self.refresh_items.add(item)
            return self.create_fake_object(item)

    def __setitem__(self, key, value):
        self.dict[key] = value
        try:
            self.refresh_items.remove(key)
        except KeyError:
            pass

    def __contains__(self, item):
        return item in self.dict

    def __len__(self):
        return len(self.dict)

    def __delitem__(self, key):
        try:
            del self.dict[key]
            self.refresh_items.remove(key)
        except KeyError:
            pass

    def values(self):
        return self.dict.values()


class VlanCache(Cache):
    object_type = Vlan
    object_key = 'number'


class InterfaceCache(Cache):
    object_type = Interface
    object_key = 'name'


class VlanInterfaceCache(Cache):
    object_type = str
    object_key = 'number'


class BondCache(Cache):
    object_type = Bond
    object_key = 'number'


class CachedSwitch(SwitchBase):
    def __init__(self, real_switch):
        super(CachedSwitch, self).__init__(real_switch.switch_descriptor)
        self.real_switch = real_switch
        self.vlans_cache = VlanCache().invalidated()
        self.interfaces_cache = InterfaceCache().invalidated()
        self.vlan_interfaces_cache = VlanInterfaceCache().invalidated()
        self.bonds_cache = BondCache().invalidated()
        self.versions_cache = Cache().invalidated()

    def _connect(self):
        return self.real_switch.connect()

    def _disconnect(self):
        return self.real_switch.disconnect()

    def _start_transaction(self):
        return self.real_switch.start_transaction()

    def commit_transaction(self):
        return self.real_switch.commit_transaction()

    def rollback_transaction(self):
        return self.real_switch.rollback_transaction()

    def _end_transaction(self):
        return self.real_switch.end_transaction()

    def get_vlan(self, number):
        if (self.vlans_cache.refresh_items and number not in self.vlans_cache) \
                or number in self.vlans_cache.refresh_items:
            self.vlans_cache[number] = self.real_switch.get_vlan(number)
        return copy.deepcopy(self.vlans_cache[number])

    def get_vlans(self):
        if None in self.vlans_cache.refresh_items:
            self.vlans_cache = VlanCache((vlan.number, vlan) for vlan in self.real_switch.get_vlans())

        for number in list(self.vlans_cache.refresh_items):
            self.get_vlan(number)

        return copy.deepcopy(self.vlans_cache.values())

    def get_vlan_interfaces(self, number):
        if (self.vlan_interfaces_cache.refresh_items and number not in self.vlan_interfaces_cache) \
                or number in self.vlan_interfaces_cache.refresh_items:
            self.vlan_interfaces_cache[number] = self.real_switch.get_vlan_interfaces(number)
        return copy.deepcopy(self.vlan_interfaces_cache[number])

    def get_interface(self, instance_id):
        if (self.interfaces_cache.refresh_items and instance_id not in self.interfaces_cache) \
                or instance_id in self.interfaces_cache.refresh_items:
            self.interfaces_cache[instance_id] = self.real_switch.get_interface(instance_id)
        return copy.deepcopy(self.interfaces_cache[instance_id])

    def get_interfaces(self):
        if self.interfaces_cache.refresh_items:
            self.interfaces_cache = InterfaceCache(
                (interface.name, interface)
                for interface in self.real_switch.get_interfaces())
        return copy.deepcopy(self.interfaces_cache.values())

    def get_bond(self, number):
        if (self.bonds_cache.refresh_items and number not in self.bonds_cache)\
                or number in self.bonds_cache.refresh_items:
            self.bonds_cache[number] = self.real_switch.get_bond(number)
        return copy.deepcopy(self.bonds_cache[number])

    def get_bonds(self):
        if self.bonds_cache.refresh_items:
            self.bonds_cache = BondCache(
                (bond.number, bond) for bond in self.real_switch.get_bonds())
        return copy.deepcopy(self.bonds_cache.values())

    def add_vlan(self, number, name=None):
        extras = {}
        if name is not None:
            extras["name"] = name
        result = self.real_switch.add_vlan(number, **extras)
        self.vlans_cache.refresh_items.add(number)
        return result

    def remove_vlan(self, number):
        self.real_switch.remove_vlan(number)
        del self.vlans_cache[number]

    def set_vlan_access_group(self, vlan_number, direction, name):
        self.real_switch.set_vlan_access_group(vlan_number, direction, name)
        self.vlans_cache[vlan_number].access_groups[direction] = name

    def unset_vlan_access_group(self, vlan_number, direction):
        self.real_switch.unset_vlan_access_group(vlan_number, direction)
        self.vlans_cache[vlan_number].access_groups[direction] = None

    def add_ip_to_vlan(self, vlan_number, ip_network):
        self.real_switch.add_ip_to_vlan(vlan_number, ip_network)
        self.vlans_cache[vlan_number].ips.append(ip_network)

    def remove_ip_from_vlan(self, vlan_number, ip_network):
        self.real_switch.remove_ip_from_vlan(vlan_number, ip_network)
        self.vlans_cache[vlan_number].ips = [
            net for net in self.vlans_cache[vlan_number].ips
            if str(net) != str(ip_network)]

    def set_vlan_vrf(self, vlan_number, vrf_name):
        self.real_switch.set_vlan_vrf(vlan_number, vrf_name)
        self.vlans_cache[vlan_number].vrf_forwarding = vrf_name

    def unset_vlan_vrf(self, vlan_number):
        self.real_switch.unset_vlan_vrf(vlan_number)
        self.vlans_cache[vlan_number].vrf_forwarding = None

    def set_access_mode(self, interface_id):
        self.real_switch.set_access_mode(interface_id)
        self.interfaces_cache[interface_id].port_mode = ACCESS
        self.interfaces_cache[interface_id].trunk_native_vlan = None
        self.interfaces_cache[interface_id].trunk_vlans = []

    def set_trunk_mode(self, interface_id):
        self.real_switch.set_trunk_mode(interface_id)
        self.interfaces_cache[interface_id].port_mode = TRUNK

    def set_bond_access_mode(self, bond_number):
        self.real_switch.set_bond_access_mode(bond_number)
        self.bonds_cache[bond_number].port_mode = ACCESS

    def set_bond_trunk_mode(self, bond_number):
        self.real_switch.set_bond_trunk_mode(bond_number)
        self.bonds_cache[bond_number].port_mode = TRUNK

    def set_access_vlan(self, interface_id, vlan):
        self.real_switch.set_access_vlan(interface_id, vlan)
        self.interfaces_cache[interface_id].access_vlan = vlan

    def reset_interface(self, interface_id):
        self.real_switch.reset_interface(interface_id)
        self.interfaces_cache.refresh_items.add(interface_id)

    def unset_interface_access_vlan(self, interface_id):
        self.real_switch.unset_interface_access_vlan(interface_id)
        self.interfaces_cache[interface_id].access_vlan = None

    def set_interface_native_vlan(self, interface_id, vlan):
        self.real_switch.set_interface_native_vlan(interface_id, vlan)
        self.interfaces_cache[interface_id].trunk_native_vlan = vlan

    def unset_interface_native_vlan(self, interface_id):
        self.real_switch.unset_interface_native_vlan(interface_id)
        self.interfaces_cache[interface_id].trunk_native_vlan = None

    def set_bond_native_vlan(self, bond_number, vlan):
        self.real_switch.set_bond_native_vlan(bond_number, vlan)
        self.bonds_cache[bond_number].trunk_native_vlan = vlan

    def unset_bond_native_vlan(self, bond_number):
        self.real_switch.unset_bond_native_vlan(bond_number)
        self.bonds_cache[bond_number].trunk_native_vlan = None

    def add_trunk_vlan(self, interface_id, vlan):
        self.real_switch.add_trunk_vlan(interface_id, vlan)
        self.interfaces_cache[interface_id].trunk_vlans.append(vlan)

    def remove_trunk_vlan(self, interface_id, vlan):
        self.real_switch.remove_trunk_vlan(interface_id, vlan)
        try:
            self.interfaces_cache[interface_id].trunk_vlans.remove(vlan)
        except ValueError:
            pass

    def add_bond_trunk_vlan(self, bond_number, vlan):
        self.real_switch.add_bond_trunk_vlan(bond_number, vlan)
        self.bonds_cache[bond_number].trunk_vlans.append(vlan)

    def remove_bond_trunk_vlan(self, bond_number, vlan):
        self.real_switch.remove_bond_trunk_vlan(bond_number, vlan)
        try:
            self.bonds_cache[bond_number].trunk_vlans.remove(vlan)
        except ValueError:
            pass

    def set_interface_description(self, interface_id, description):
        # No cache to update
        self.real_switch.set_interface_description(interface_id, description)

    def unset_interface_description(self, interface_id):
        # No cache to update
        self.real_switch.unset_interface_description(interface_id)

    def set_bond_description(self, bond_number, description):
        # No cache to update
        self.real_switch.set_bond_description(bond_number, description)

    def unset_bond_description(self, bond_number):
        # No cache to update
        self.real_switch.unset_bond_description(bond_number)

    def edit_interface_spanning_tree(self, interface_id, edge=None):
        # No cache to update
        self.real_switch.edit_interface_spanning_tree(interface_id, edge=edge)

    def set_interface_state(self, interface_id, state):
        self.real_switch.set_interface_state(interface_id, state)
        self.interfaces_cache[interface_id].shutdown = (state == OFF)

    def unset_interface_state(self, interface_id):
        self.real_switch.unset_interface_state(interface_id)
        self.interfaces_cache.refresh_items.add(interface_id)

    def set_interface_auto_negotiation_state(self, interface_id, state):
        self.real_switch.set_interface_auto_negotiation_state(interface_id, state)
        self.interfaces_cache[interface_id].auto_negotiation = (state == ON)

    def unset_interface_auto_negotiation_state(self, interface_id):
        self.real_switch.unset_interface_auto_negotiation_state(interface_id)
        self.interfaces_cache[interface_id].auto_negotiation = None

    def add_bond(self, number):
        self.real_switch.add_bond(number)
        self.bonds_cache.refresh_items.add(number)

    def remove_bond(self, number):
        self.real_switch.remove_bond(number)
        del self.bonds_cache[number]

    def add_interface_to_bond(self, interface, bond_number):
        self.real_switch.add_interface_to_bond(interface, bond_number)
        self.bonds_cache[bond_number].members.append(interface)
        self.interfaces_cache.refresh_items.add(interface)

    def remove_interface_from_bond(self, interface):
        self.real_switch.remove_interface_from_bond(interface)
        self.interfaces_cache[interface].bond_master = None
        self.interfaces_cache.refresh_items.add(interface)
        for bond in self.bonds_cache.values():
            try:
                bond.members.remove(interface)
            except ValueError:
                pass

    def set_bond_link_speed(self, number, speed):
        self.real_switch.set_bond_link_speed(number, speed)
        self.bonds_cache[number].link_speed = speed

    def edit_bond_spanning_tree(self, number, edge=None):
        self.real_switch.edit_bond_spanning_tree(number, edge=edge)

    def add_vrrp_group(self, vlan_number, group_id, ips=None, priority=None,
                       hello_interval=None, dead_interval=None, track_id=None,
                       track_decrement=None):
        self.real_switch.add_vrrp_group(vlan_number, group_id, ips=ips,
                                        priority=priority,
                                        hello_interval=hello_interval,
                                        dead_interval=dead_interval,
                                        track_id=track_id,
                                        track_decrement=track_decrement)
        self.vlans_cache[vlan_number].vrrp_groups.append(VrrpGroup(
            id=group_id, ips=ips, priority=priority,
            hello_interval=hello_interval, dead_interval=dead_interval,
            track_id=track_id, track_decrement=track_decrement
        ))

    def remove_vrrp_group(self, vlan_number, group_id):
        self.real_switch.remove_vrrp_group(vlan_number, group_id)
        for group in self.vlans_cache[vlan_number].vrrp_groups:
            if group.id == group_id:
                self.vlans_cache[vlan_number].vrrp_groups.remove(group)

    def add_vlan_varp_ip(self, vlan_number, ip_network):
        self.real_switch.add_vlan_varp_ip(vlan_number, ip_network)
        self.vlans_cache[vlan_number].varp_ips.append(ip_network)

    def remove_vlan_varp_ip(self, vlan_number, ip_network):
        self.real_switch.remove_vlan_varp_ip(vlan_number, ip_network)
        self.vlans_cache[vlan_number].varp_ips.remove(ip_network)

    def add_dhcp_relay_server(self, vlan_number, ip_address):
        self.real_switch.add_dhcp_relay_server(vlan_number, ip_address)
        self.vlans_cache[vlan_number].dhcp_relay_servers.append(ip_address)

    def remove_dhcp_relay_server(self, vlan_number, ip_address):
        self.real_switch.remove_dhcp_relay_server(vlan_number, ip_address)
        try:
            self.vlans_cache[vlan_number].dhcp_relay_servers.remove(ip_address)
        except ValueError:
            pass

    def set_interface_lldp_state(self, interface_id, enabled):
        self.real_switch.set_interface_lldp_state(interface_id, enabled)

    def set_vlan_icmp_redirects_state(self, vlan_number, state):
        self.real_switch.set_vlan_icmp_redirects_state(vlan_number, state)
        self.vlans_cache[vlan_number].icmp_redirects = state

    def set_vlan_ntp_state(self, vlan_number, state):
        self.real_switch.set_vlan_ntp_state(vlan_number, state)
        self.vlans_cache[vlan_number].ntp = state

    def set_vlan_unicast_rpf_mode(self, vlan_number, mode):
        self.real_switch.set_vlan_unicast_rpf_mode(vlan_number, mode)
        self.vlans_cache[vlan_number].unicast_rpf_mode = mode

    def unset_vlan_unicast_rpf_mode(self, vlan_number):
        self.real_switch.unset_vlan_unicast_rpf_mode(vlan_number)
        self.vlans_cache[vlan_number].unicast_rpf_mode = None

    def get_versions(self):
        if self.versions_cache.refresh_items:
            self.versions_cache = Cache([(0, self.real_switch.get_versions())])
        return copy.deepcopy(self.versions_cache[0])

    def set_interface_mtu(self, interface_id, size):
        self.real_switch.set_interface_mtu(interface_id, size)
        self.interfaces_cache[interface_id].mtu = size

    def unset_interface_mtu(self, interface_id):
        self.real_switch.unset_interface_mtu(interface_id)
        self.interfaces_cache[interface_id].mtu = None

    def set_bond_mtu(self, bond_number, size):
        self.real_switch.set_bond_mtu(bond_number, size)
        self.bonds_cache[bond_number].mtu = size

    def unset_bond_mtu(self, bond_number):
        self.real_switch.unset_bond_mtu(bond_number)
        self.bonds_cache[bond_number].mtu = None

    def set_vlan_arp_routing_state(self, vlan_number, state):
        self.real_switch.set_vlan_arp_routing_state(vlan_number, state)
        self.vlans_cache[vlan_number].arp_routing = (state == ON)



from hamcrest import assert_that, equal_to
from netman.core.objects.exceptions import UnknownBond
from tests.adapters.compliance_test_case import ComplianceTestCase


class GetBondTest(ComplianceTestCase):
    _dev_sample = "juniper"

    def setUp(self):
        super(GetBondTest, self).setUp()
        self.client.add_bond(42)
        self.client.add_vlan(99)

    def tearDown(self):
        self.janitor.remove_bond(42)
        self.janitor.remove_vlan(99)
        super(GetBondTest, self).tearDown()

    def test_get_bond_and_get_bonds_return_the_same_for_sport_speed(self):
        self.try_to.set_bond_link_speed(42, '1g')

        from_get_bond, from_get_bonds = self._get_bond_results(42)

        assert_that(from_get_bond.link_speed, equal_to(from_get_bonds.link_speed))

    def test_get_bond_and_get_bonds_return_the_same_on_access_mode(self):
        self.try_to.set_bond_access_mode(42)

        from_get_bond, from_get_bonds = self._get_bond_results(42)

        assert_that(from_get_bond.port_mode, equal_to(from_get_bonds.port_mode))

    def test_get_bond_and_get_bonds_return_the_same_on_trunk_mode(self):
        self.try_to.add_vlan(100)
        self.try_to.set_bond_trunk_mode(42)
        self.try_to.set_bond_native_vlan(42, 99)
        self.try_to.add_bond_trunk_vlan(42, 100)

        from_get_bond, from_get_bonds = self._get_bond_results(42)

        assert_that(from_get_bond.port_mode, equal_to(from_get_bonds.port_mode))
        assert_that(from_get_bond.trunk_vlans, equal_to(from_get_bonds.trunk_vlans))
        assert_that(from_get_bond.trunk_native_vlan, equal_to(from_get_bonds.trunk_native_vlan))

        self.janitor.remove_vlan(100)

    def _get_bond_results(self, bond_number):
        return self.client.get_bond(bond_number), \
               [bond for bond in self.client.get_bonds() if bond.number == bond_number][0]

    def test_raises_when_bond_does_not_exist(self):
        with self.assertRaises(UnknownBond):
            self.client.get_bond(999)



from hamcrest import is_, assert_that, empty, none
from netman.core.objects.exceptions import UnknownInterface
from netman.core.objects.port_modes import BOND_MEMBER
from tests.adapters.compliance_test_case import ComplianceTestCase


class AddInterfaceToBondTest(ComplianceTestCase):
    _dev_sample = "juniper"

    def setUp(self):
        super(AddInterfaceToBondTest, self).setUp()
        self.client.add_bond(42)

    def tearDown(self):
        self.janitor.remove_bond(42)
        super(AddInterfaceToBondTest, self).tearDown()

    def test_sets_the_interface_bond_master(self):
        self.client.add_interface_to_bond(self.test_port, 42)
        interface = self.client.get_interface(self.test_port)

        assert_that(interface.bond_master, is_(42))

        self.janitor.remove_interface_from_bond(self.test_port)

    def test_sets_the_interface_port_mode_to_bond_members(self):
        self.client.add_interface_to_bond(self.test_port, 42)
        interface = self.client.get_interface(self.test_port)

        assert_that(interface.port_mode, is_(BOND_MEMBER))

        self.janitor.remove_interface_from_bond(self.test_port)

    def test_raises_on_invalid_interface(self):
        with self.assertRaises(UnknownInterface):
            self.client.add_interface_to_bond("nonexistent 1/42", 42)

    def test_resets_the_interface_trunk_vlans(self):
        self.client.add_vlan(90)
        self.client.set_trunk_mode(self.test_port)
        self.client.add_trunk_vlan(self.test_port, 90)

        self.client.add_interface_to_bond(self.test_port, 42)
        interface = self.client.get_interface(self.test_port)

        assert_that(interface.trunk_vlans, is_(empty()))
        self.janitor.remove_vlan(90)
        self.janitor.remove_interface_from_bond(self.test_port)
        self.janitor.set_access_mode(self.test_port)

    def test_resets_the_interface_trunk_native_vlan(self):
        self.client.add_vlan(90)
        self.client.set_trunk_mode(self.test_port)
        self.client.set_interface_native_vlan(self.test_port, 90)

        self.client.add_interface_to_bond(self.test_port, 42)
        interface = self.client.get_interface(self.test_port)

        assert_that(interface.trunk_native_vlan, is_(none()))
        self.janitor.remove_vlan(90)
        self.janitor.remove_interface_from_bond(self.test_port)
        self.janitor.set_access_mode(self.test_port)

    def test_resets_the_interface_access_vlan(self):
        self.client.add_vlan(90)
        self.client.set_access_mode(self.test_port)
        self.client.set_access_vlan(self.test_port, 90)

        self.client.add_interface_to_bond(self.test_port, 42)
        interface = self.client.get_interface(self.test_port)

        assert_that(interface.access_vlan, is_(none()))
        self.janitor.remove_vlan(90)



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import traceback

from netman import regex
import re


class SubShell(object):
    debug = False

    def __init__(self, ssh, enter, exit_cmd, validate=None):
        self.ssh = ssh
        self.enter = enter
        self.exit = exit_cmd
        self.validate = validate or (lambda x: None)

    def __enter__(self):
        if isinstance(self.enter, list):
            [self.validate(self.ssh.do(cmd)) for cmd in self.enter]
        else:
            self.validate(self.ssh.do(self.enter))
        return self.ssh

    def __exit__(self, eType, eValue, eTrace):
        if self.debug and eType is not None:
            logging.error("Subshell exception {}: {}\n{}"
                          .format(eType.__name__, eValue, "".join(traceback.format_tb(eTrace))))

        self.ssh.do(self.exit)


def no_output(exc, *args):
    def m(welcome_msg):
        if len(welcome_msg) > 0:
            raise exc(*args)
    return m


def split_on_bang(data):
    current_chunk = []
    for line in data:
        if re.match("^!.*", line):
            if len(current_chunk) > 0:
                yield current_chunk
                current_chunk = []
        else:
            current_chunk.append(line)


def split_on_dedent(data):
    current_chunk = []
    for line in data:
        if re.match("^[^\s].*", line) and len(current_chunk) > 0:
            yield current_chunk
            current_chunk = [line]
        else:
            current_chunk.append(line)

    yield current_chunk


class ResultChecker(object):
    def __init__(self, result=None):
        self.result = result

    def on_any_result(self, exception, *args, **kwargs):
        if self.result and len(self.result) > 0:
            raise exception(*args, **kwargs)
        return self

    def on_result_matching(self, matcher, exception, *args, **kwargs):
        if regex.match(matcher, "\n".join(self.result), flags=re.DOTALL):
            raise exception(*args, **kwargs)
        return self


class PageReader(object):
    def __init__(self, read_while, and_press, unless_prompt):
        self.next_page_indicator = read_while
        self.continue_key = and_press
        self.prompt = unless_prompt

    def do(self, shell, command):
        result = shell.do(command,
                          wait_for=(self.next_page_indicator, self.prompt),
                          include_last_line=True)

        while len(result) > 0 and self.next_page_indicator in result[-1]:
            result = result[:-1] + shell.send_key(self.continue_key,
                                                  wait_for=(self.next_page_indicator, self.prompt),
                                                  include_last_line=True)

        return result[:-1]



import unittest
from copy import deepcopy

from hamcrest import assert_that
from hamcrest import is_

from netman.adapters.threading_lock_factory import ThreadingLockFactory
from netman.core.switch_factory import FlowControlSwitchFactory
from tests.system_tests import NetmanTestApp, get_available_switch


class SessionTest(unittest.TestCase):

    def test_multiple_proxies_works(self):
        with NetmanTestApp() as partial_client1, NetmanTestApp() as partial_client2:
            factory = FlowControlSwitchFactory(None, ThreadingLockFactory())
            switch_descriptor = deepcopy(get_available_switch("juniper"))

            client1 = partial_client1(switch_descriptor)
            first_netman_url = "{}:{}".format(client1.host, client1.port)

            client2 = partial_client2(switch_descriptor)
            second_netman_url = "{}:{}".format(client2.host, client2.port)

            switch_descriptor.netman_server = [first_netman_url, second_netman_url]

            switch = factory.get_switch_by_descriptor(switch_descriptor)

            with switch.transaction():
                switch.add_vlan(1497, "one")

            assert_that(client2.get("/switches/{hostname}/vlans/1497").json()["name"], is_("one"))



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import types
from contextlib import contextmanager
from functools import wraps

from netman.core.objects.switch_base import SwitchOperations


def do_not_wrap_with_flow_control(fn):
    setattr(fn, "_do_not_wrap_with_flow_control", True)
    return fn


class FlowControlSwitch(SwitchOperations):
    """
    Wrap your switch with this to handle auto-connections and auto-transactions

    fc_switch = FlowControlSwitch(myswitch)

    with fc_switch.transaction(): #will lock, connect and transaction
        fc_switch.add_vlan(1000)

    OR

    fc_switch.start_transaction() # will auto connect
    fc_switch.add_vlan()
    fc_switch.end_transaction() # will auto disconnect since it auto connected

    OR

    fc_switch.add_vlan(1000) #will auto lock, connect and transaction

    """
    def __init__(self, wrapped_switch, lock):
        self.wrapped_switch = wrapped_switch
        self.lock = lock
        self._has_auto_connected = False

    def __new__(cls, *args, **kwargs):
        obj = super(FlowControlSwitch, cls).__new__(cls, *args, **kwargs)

        for member in dir(cls):
            if not member.startswith("_"):
                _wrap_method_with_flow_control(cls, obj, member)

        return obj

    @do_not_wrap_with_flow_control
    @contextmanager
    def transaction(self):
        with self._locked_context(), self._connected_context(), self._transaction_context():
            yield

    @do_not_wrap_with_flow_control
    def start_transaction(self):
        if self.wrapped_switch.in_transaction:
            return

        self.lock.acquire()
        try:
            if not self.wrapped_switch.connected:
                self.wrapped_switch.connect()
                self._has_auto_connected = True

            self.wrapped_switch.start_transaction()
        except Exception:
            try:
                if self._has_auto_connected:
                    self.wrapped_switch.disconnect()
                    self._has_auto_connected = False
            finally:
                self.lock.release()
            raise

    @do_not_wrap_with_flow_control
    def end_transaction(self):
        try:
            self.wrapped_switch.end_transaction()
        finally:
            try:
                if self._has_auto_connected:
                    self.wrapped_switch.disconnect()
                    self._has_auto_connected = False
            finally:
                self.lock.release()

    @contextmanager
    def _connected_context(self):
        if self.wrapped_switch.connected:
            yield
        else:
            self.wrapped_switch.connect()
            try:
                yield
            finally:
                self.wrapped_switch.disconnect()

    @contextmanager
    def _transaction_context(self):
        if self.wrapped_switch.in_transaction:
            yield
        else:
            self.wrapped_switch.start_transaction()
            try:
                yield
                self.wrapped_switch.commit_transaction()
            except Exception:
                self.wrapped_switch.rollback_transaction()
                raise
            finally:
                self.wrapped_switch.end_transaction()

    @contextmanager
    def _locked_context(self):
        if self.wrapped_switch.in_transaction:
            yield
        else:
            self.lock.acquire()
            try:
                yield
            finally:
                self.lock.release()

    @do_not_wrap_with_flow_control
    def connect(self):
        self.wrapped_switch.connect()

    @do_not_wrap_with_flow_control
    def disconnect(self):
        self.wrapped_switch.disconnect()

    @do_not_wrap_with_flow_control
    def commit_transaction(self):
        self.wrapped_switch.commit_transaction()

    @do_not_wrap_with_flow_control
    def rollback_transaction(self):
        self.wrapped_switch.rollback_transaction()

    @property
    def switch_descriptor(self):
        return self.wrapped_switch.switch_descriptor


def _wrap_method_with_flow_control(cls, obj, method_name):
    original = getattr(cls, method_name)
    if not callable(original) or isinstance(original, property) or hasattr(original, "_do_not_wrap_with_flow_control"):
        return

    if method_name.startswith("get_"):
        @wraps(original)
        def wrapped(self, *args, **kwargs):
            with self._connected_context():
                return getattr(self.wrapped_switch, method_name)(*args, **kwargs)
    else:
        @wraps(original)
        def wrapped(self, *args, **kwargs):
            with self.transaction():
                return getattr(self.wrapped_switch, method_name)(*args, **kwargs)

    setattr(obj, method_name, types.MethodType(wrapped, obj))



import time
import unittest

import sys
from hamcrest import assert_that, less_than
from hamcrest import is_
import os

from tests.system_tests import NetmanTestApp, create_session, get_available_switch


class GunicornCompatibilityTest(unittest.TestCase):
    def test_run_netman_with_a_gunicorn_wrapper(self):
        with GunicornNetmanTestApp() as partial_client:
            client = partial_client(get_available_switch("cisco"))

            create_session(client, "my_session")

            result = client.delete("/switches-sessions/my_session")
            assert_that(result.status_code, is_(204), result.text)

    def test_parameters_can_be_passed_through_the_command_line(self):
        with GunicornNetmanTestApp() as partial_client:
            client = partial_client(get_available_switch("brocade"))
            start_time = time.time()

            create_session(client, "session_timeouting")

            create_session(client, "session_taking_over")

            result = client.delete("/switches-sessions/session_taking_over")
            assert_that(result.status_code, is_(204), result.text)

            assert_that(time.time() - start_time, is_(less_than(3)))


class GunicornNetmanTestApp(NetmanTestApp):
    def _popen_params(self, path):
        gunicorn_executable = os.path.join(os.path.dirname(sys.executable), 'gunicorn')
        params = [gunicorn_executable,
                  "netman.main:load_app(session_inactivity_timeout={})".format(self.session_inactivity_timeout),
                  "--bind", "{}:{}".format(self.ip, self.port),
                  "--threads", "2"]
        return params



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import unittest


class ExceptionsComplianceTest(unittest.TestCase):

    def test_for_remote_use_all_exceptions_should_be_instantiable_without_an_argument(self):

        exceptions_module = importlib.import_module("netman.core.objects.exceptions")

        for attribute in dir(exceptions_module):
            exception_class = getattr(exceptions_module, attribute)
            if isinstance(exception_class, type):
                try:
                    exception_class()
                except:
                    raise AssertionError("Class {0} should be instantiable with no params".format(attribute))



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hamcrest import assert_that, has_item, is_not, has_properties
from netman.core.objects.exceptions import UnknownBond

from tests.adapters.configured_test_case import skip_on_switches, ConfiguredTestCase


class BondManagementTest(ConfiguredTestCase):
    __test__ = False

    @skip_on_switches("cisco", "brocade", "brocade_telnet", "dell", "dell_telnet", "dell10g", "dell10g_telnet", "juniper_mx", "arista_http")
    def test_creating_deleting_a_bond(self):
        self.client.add_bond(3)

        list_of_bonds = self.client.get_bonds()
        assert_that(list_of_bonds, has_item(has_properties(number=3)))

        bond = self.client.get_bond(3)
        assert_that(bond, has_properties(number=3))

        self.client.remove_bond(3)

        new_list_of_bonds = self.client.get_bonds()
        assert_that(new_list_of_bonds, is_not(has_item(has_properties(number=3))))

        with self.assertRaises(UnknownBond):
            self.client.get_bond(3)



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from functools import wraps
from unittest import SkipTest

from hamcrest import assert_that, is_
from netman.adapters.switches.cached import CachedSwitch
from netman.adapters.switches.remote import RemoteSwitch
from netman.core.objects.exceptions import NetmanException
from netman.main import app
from tests.adapters.flask_helper import FlaskRequest
from tests.adapters.model_list import available_models


def sub_dict(d, *keys):
    return dict((k, d[k]) for k in keys)


class ValidatingCachedSwitch(CachedSwitch):
    def get_bond(self, number):
        bond = super(ValidatingCachedSwitch, self).get_bond(number)
        assert_that(bond, is_(self.real_switch.get_bond(number)))
        return bond

    def get_bonds(self):
        bonds = super(ValidatingCachedSwitch, self).get_bonds()
        assert_that(bonds, is_(self.real_switch.get_bonds()))
        return bonds

    def get_interfaces(self):
        interfaces = super(ValidatingCachedSwitch, self).get_interfaces()
        assert_that(interfaces, is_(self.real_switch.get_interfaces()))
        return interfaces

    def get_interface(self, interface_id):
        interface = super(ValidatingCachedSwitch, self).get_interface(interface_id)
        assert_that(interface, is_(self.real_switch.get_interface(interface_id)))
        return interface

    def get_vlan(self, number):
        vlan = super(ValidatingCachedSwitch, self).get_vlan(number)
        assert_that(vlan, is_(self.real_switch.get_vlan(number)))
        return vlan

    def get_vlans(self):
        vlans = super(ValidatingCachedSwitch, self).get_vlans()
        assert_that(vlans, is_(self.real_switch.get_vlans()))
        return vlans

    def get_vlan_interfaces(self, number):
        vlan_interfaces = super(ValidatingCachedSwitch, self).get_vlan_interfaces(number)
        assert_that(vlan_interfaces, is_(self.real_switch.get_vlan_interfaces(number)))
        return vlan_interfaces


class ConfiguredTestCase(unittest.TestCase):
    _dev_sample = None
    switch_specs = None

    def setUp(self):
        if self.switch_specs is not None:
            specs = type(self).switch_specs
        else:
            specs = next(s for s in available_models if s["switch_descriptor"].model == self._dev_sample)

        self.switch_descriptor = specs["switch_descriptor"]
        self.test_port = specs["test_port_name"]
        self.test_ports = specs["ports"]
        self.test_vrrp_track_id = specs.get("test_vrrp_track_id")

        self.switch_descriptor.netman_server = ''
        self.remote_switch = RemoteSwitch(self.switch_descriptor)
        self.remote_switch.requests = FlaskRequest(app.test_client())

        self.client = ValidatingCachedSwitch(self.remote_switch)
        self.try_to = ExceptionIgnoringProxy(self.client, [NotImplementedError])
        self.janitor = ExceptionIgnoringProxy(self.client, [NotImplementedError, NetmanException])

        self.client.connect()
        self.client.start_transaction()

    def tearDown(self):
        self.client.end_transaction()
        self.client.disconnect()

    def get_vlan_from_list(self, number):
        try:
            return next((vlan for vlan in self.client.get_vlans()
                         if vlan.number == number))
        except StopIteration:
            raise AssertionError("Vlan #{} not found".format(number))


def skip_on_switches(*to_skip):

    def resource_decorator(fn):
        @wraps(fn)
        def wrapper(self, *args, **kwargs):
            if self.switch_descriptor.model not in to_skip:
                return fn(self, *args, **kwargs)

            else:
                raise SkipTest('Test not executed on Switch model %s' % self.switch_descriptor.model)

        return wrapper

    return resource_decorator


class ExceptionIgnoringProxy(object):
    def __init__(self, target, exceptions):
        self.target = target
        self.exceptions = tuple(exceptions)

    def __getattr__(self, item):
        def wrapper(*args, **kwargs):
            try:
                return getattr(self.target, item)(*args, **kwargs)
            except self.exceptions:
                return None

        return wrapper



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License

import time

from MockSSH import SSHCommand


class HangingCommand(SSHCommand):
    def __init__(self, name, hang_time, *args):
        self.name = name
        self.hang_time = hang_time
        self.protocol = None  # set in __call__

    def __call__(self, protocol, *args):
        SSHCommand.__init__(self, protocol, self.name, *args)
        return self

    def start(self):
        time.sleep(self.hang_time)
        self.write("Done!\n")
        self.exit()


class AmbiguousCommand(SSHCommand):
    def __init__(self, name, *args):
        self.name = name
        self.protocol = None  # set in __call__

    def __call__(self, protocol, *args):
        SSHCommand.__init__(self, protocol, self.name, *args)
        return self

    def start(self):
        self.write("working -> done!\n")
        self.exit()


class MultiAsyncWriteCommand(SSHCommand):
    def __init__(self, name, count, interval, *args):
        self.name = name
        self.count = count
        self.interval = interval
        self.protocol = None  # set in __call__

    def __call__(self, protocol, *args):
        SSHCommand.__init__(self, protocol, self.name, *args)
        return self

    def start(self):
        for i in range(self.count):
            self.write("Line %d\n" % (i + 1))
            time.sleep(self.interval)

        self.exit()


class SkippingLineCommand(SSHCommand):
    def __init__(self, name, lines, *args):
        self.name = name
        self.lines = lines
        self.protocol = None  # set in __call__

    def __call__(self, protocol, *args):
        SSHCommand.__init__(self, protocol, self.name, *args)
        return self

    def start(self):
        for _ in range(self.lines):
            self.write("\r\n")

        self.write("%s lines skipped!\n" % self.lines)
        self.exit()


def exit_command_success(instance):
    instance.protocol.call_command(instance.protocol.commands['_exit'])


def passwd_change_protocol_prompt(instance):
    instance.protocol.prompt = "hostname#"
    instance.protocol.password_input = False


def passwd_write_password_to_transport(instance):
    instance.writeln("MockSSH: password is %s" % instance.valid_password)


class KeystrokeAnsweredCommand(SSHCommand):
    def __init__(self, name):
        self.name = name
        self.protocol = None  # set in __call__

    def __call__(self, protocol, *args):
        SSHCommand.__init__(self, protocol, self.name, *args)
        return self

    def start(self):
        self.write("whatup?")

        this = self

        def finish():
            this.writeln("k")
            this.writeln("K pressed")
            this.exit()
            this.protocol.keyHandlers.pop("k")

        self.protocol.keyHandlers.update({"k": finish})



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import warnings

from netman.adapters.shell.ssh import SshClient
from netman.adapters.shell.telnet import TelnetClient
from netman.core.objects.interface_states import OFF
from netman.core.objects.port_modes import TRUNK
from netman.core.objects.port_modes import ACCESS
from netman.core.objects.interface import Interface
from netman.adapters.switches.cisco import parse_vlan_ranges
from netman.core.objects.vlan import Vlan
from netman import regex
from netman.core.objects.switch_transactional import FlowControlSwitch
from netman.adapters.switches.util import SubShell, no_output, ResultChecker, PageReader
from netman.core.objects.exceptions import UnknownInterface, BadVlanName, \
    BadVlanNumber, UnknownVlan, InterfaceInWrongPortMode, NativeVlanNotSet, TrunkVlanNotSet, BadInterfaceDescription, \
    VlanAlreadyExist, UnknownBond, InvalidMtuSize, InterfaceResetIncomplete, \
    PrivilegedAccessRefused
from netman.core.objects.switch_base import SwitchBase


def ssh(switch_descriptor):
    return Dell(switch_descriptor, shell_factory=SshClient)


def telnet(switch_descriptor):
    return Dell(switch_descriptor, shell_factory=TelnetClient)


def factory_ssh(switch_descriptor, lock):
    warnings.warn("Use SwitchFactory.get_switch_by_descriptor directly to instantiate a switch", DeprecationWarning)
    return FlowControlSwitch(wrapped_switch=ssh(switch_descriptor), lock=lock)


def factory_telnet(switch_descriptor, lock):
    warnings.warn("Use SwitchFactory.get_switch_by_descriptor directly to instantiate a switch", DeprecationWarning)
    return FlowControlSwitch(wrapped_switch=telnet(switch_descriptor), lock=lock)


class Dell(SwitchBase):

    def __init__(self, switch_descriptor, shell_factory):
        super(Dell, self).__init__(switch_descriptor)
        self.shell = None
        self.shell_factory = shell_factory

        self.page_reader = PageReader(
            read_while="--More-- or (q)uit",
            and_press="m",
            unless_prompt="#"
        )

    def _connect(self):
        params = dict(
            host=self.switch_descriptor.hostname,
            username=self.switch_descriptor.username,
            password=self.switch_descriptor.password,
        )

        if self.switch_descriptor.port:
            params["port"] = self.switch_descriptor.port

        self.shell = self.shell_factory(**params)

        self.shell.do("enable", wait_for=":")
        password_return = self.shell.do(self.switch_descriptor.password)
        if any(["Incorrect Password" in line for line in password_return]):
            raise PrivilegedAccessRefused(password_return)

    def _disconnect(self):
        self.shell.quit("quit")
        self.logger.info(self.shell.full_log)

    def _start_transaction(self):
        pass

    def _end_transaction(self):
        pass

    def rollback_transaction(self):
        pass

    def commit_transaction(self):
        self.shell.do("copy running-config startup-config", wait_for="? (y/n) ")
        self.shell.send_key("y")

    def set_interface_state(self, interface_id, state):
        with self.config(), self.interface(interface_id):
            self.shell.do('shutdown' if state is OFF else 'no shutdown')

    def get_vlans(self):
        result = self.page_reader.do(self.shell, "show vlan")
        vlans = parse_vlan_list(result)
        return vlans

    def get_vlan(self, vlan_number):
        result = self.page_reader.do(self.shell, "show vlan id {}".format(vlan_number))

        if regex.match(".*\^.*", result[0]):
            raise BadVlanNumber()
        elif regex.match("^ERROR", result[0]):
            raise UnknownVlan(vlan_number)
        vlan = parse_vlan_list(result)[0]
        return vlan

    def get_vlan_interfaces(self, vlan_number):
        result = self.page_reader.do(self.shell, "show vlan id {}".format(vlan_number))

        if regex.match(".*\^.*", result[0]):
            raise BadVlanNumber()
        elif regex.match("^ERROR", result[0]):
            raise UnknownVlan(vlan_number)

        return self.parse_interface_from_vlan_list(vlan_number, result)

    def get_interface(self, interface_id):
        return self.read_interface(interface_id)

    def get_interfaces(self):
        result = self.page_reader.do(self.shell, 'show interfaces status')

        return [self.read_interface(name) for name in self.parse_interface_names(result)]

    def add_vlan(self, number, name=None):
        result = self.page_reader.do(self.shell, "show vlan id {}".format(number))
        if regex.match(".*\^.*", result[0]):
            raise BadVlanNumber()
        elif regex.match("^VLAN", result[0]):
            raise VlanAlreadyExist(number)

        with self.config():
            with self.vlan_database():
                self.set('vlan {}', number)

            if name is not None:
                with self.interface("vlan {}".format(number)):
                    self.set('name {}', name).on_any_result(BadVlanName)

    def remove_vlan(self, number, name=None):
        with self.config():
            with self.vlan_database():
                self.set('no vlan {}', number).on_result_matching(".*These VLANs do not exist:.*", UnknownVlan, number)

    def set_interface_description(self, interface_id, description):
        with self.config(), self.interface(interface_id):
            self.set('description "{}"', description).on_any_result(BadInterfaceDescription, description)

    def set_access_mode(self, interface_id):
        with self.config(), self.interface(interface_id):
            self.shell.do("switchport mode access")

    def set_trunk_mode(self, interface_id):
        interface_data = self.get_interface_data(interface_id)
        actual_port_mode = resolve_port_mode(interface_data)

        if actual_port_mode in ("access", None):
            with self.config(), self.interface(interface_id):
                self.shell.do("switchport mode trunk")

    def set_access_vlan(self, interface_id, vlan):
        with self.config(), self.interface(interface_id):
            self.set("switchport access vlan {}", vlan)\
                .on_result_matching(".*VLAN ID not found.*", UnknownVlan, vlan)\
                .on_result_matching(".*Interface not in Access Mode.*", InterfaceInWrongPortMode, "trunk")

    def set_interface_mtu(self, interface_id, size):
        with self.config(), self.interface(interface_id):
            self.set("mtu {}", size)\
                .on_result_matching(".*Value is out of range.*", InvalidMtuSize, size)

    def unset_interface_mtu(self, interface_id):
        with self.config(), self.interface(interface_id):
            self.set("no mtu")

    def set_bond_mtu(self, number, size):
        with NamedBond(number) as bond:
            self.set_interface_mtu(bond.name, size)

    def unset_bond_mtu(self, number):
        with NamedBond(number) as bond:
            self.unset_interface_mtu(bond.name)

    def unset_interface_access_vlan(self, interface_id):
        with self.config(), self.interface(interface_id):
            self.shell.do("no switchport access vlan")

    def set_interface_native_vlan(self, interface_id, vlan):
        interface_data = self.get_interface_data(interface_id)

        actual_port_mode = resolve_port_mode(interface_data)
        if actual_port_mode == "access":
            raise InterfaceInWrongPortMode("access")

        with self.config(), self.interface(interface_id):
            if actual_port_mode != "general":
                self.set("switchport mode general")
            if actual_port_mode == "trunk":
                self.copy_vlans(interface_data, "trunk", "general")

            self.set("switchport general pvid {}", vlan).on_any_result(UnknownVlan, vlan)

    def unset_interface_native_vlan(self, interface_id):
        interface_data = self.get_interface_data(interface_id)
        assert_native_vlan_is_set(interface_id, interface_data)

        with self.config(), self.interface(interface_id):
            self.set("switchport mode trunk")
            self.copy_vlans(interface_data, "general", "trunk")

    def add_trunk_vlan(self, interface_id, vlan):
        interface_data = self.get_interface_data(interface_id)

        actual_port_mode = resolve_port_mode(interface_data)
        if actual_port_mode == "access":
            raise InterfaceInWrongPortMode("access")

        with self.config(), self.interface(interface_id):
            if actual_port_mode is None:
                self.set("switchport mode trunk")
                actual_port_mode = "trunk"

            self.set("switchport {} allowed vlan add {}", actual_port_mode, vlan)\
                .on_result_matching(".*VLAN does not exist.*", UnknownVlan, vlan)

    def remove_trunk_vlan(self, interface_id, vlan):
        interface_data = self.get_interface_data(interface_id)
        trunk_vlans = resolve_trunk_vlans(interface_data)

        if vlan not in trunk_vlans:
            raise TrunkVlanNotSet(interface_id)

        actual_port_mode = resolve_port_mode(interface_data)
        with self.config(), self.interface(interface_id):
            self.set("switchport {} allowed vlan remove {}", actual_port_mode, vlan)

    def edit_interface_spanning_tree(self, interface_id, edge=None):
        commands = []
        if edge is not None:
            commands.append("{}spanning-tree portfast".format("" if edge else "no "))

        if commands:
            with self.config(), self.interface(interface_id):
                [self.shell.do(cmd) for cmd in commands]

    def set_interface_lldp_state(self, interface_id, enabled):
        with self.config(), self.interface(interface_id):
            self.set("{}lldp transmit", "" if enabled else "no ")
            self.set("{}lldp receive", "" if enabled else "no ")
            self.set("{}lldp med transmit-tlv capabilities", "" if enabled else "no ")
            self.set("{}lldp med transmit-tlv network-policy", "" if enabled else "no ")

    def set_bond_description(self, number, description):
        with NamedBond(number) as bond:
            return self.set_interface_description(bond.name, description)

    def set_bond_trunk_mode(self, number):
        with NamedBond(number) as bond:
            return self.set_trunk_mode(bond.name)

    def set_bond_access_mode(self, number):
        with NamedBond(number) as bond:
            return self.set_access_mode(bond.name)

    def add_bond_trunk_vlan(self, number, vlan):
        with NamedBond(number) as bond:
            return self.add_trunk_vlan(bond.name, vlan)

    def remove_bond_trunk_vlan(self, number, vlan):
        with NamedBond(number) as bond:
            return self.remove_trunk_vlan(bond.name, vlan)

    def set_bond_native_vlan(self, number, vlan):
        with NamedBond(number) as bond:
            return self.set_interface_native_vlan(bond.name, vlan)

    def unset_bond_native_vlan(self, number):
        with NamedBond(number) as bond:
            return self.unset_interface_native_vlan(bond.name)

    def config(self):
        return SubShell(self.shell, enter="configure", exit_cmd='exit')

    def vlan_database(self):
        return SubShell(self.shell, enter="vlan database", exit_cmd='exit')

    def interface(self, interface_id):
        return SubShell(self.shell, enter="interface {}".format(interface_id), exit_cmd='exit',
                        validate=no_output(UnknownInterface, interface_id))

    def set(self, command, *arguments):
        result = self.shell.do(command.format(*arguments))

        return ResultChecker(result)

    def get_interface_data(self, interface_id):
        interface_data = self.page_reader.do(self.shell, "show running-config interface {}".format(interface_id))
        if any(["Invalid input" in line or regex.match("ERROR.*", line) for line in interface_data]):
            raise UnknownInterface(interface_id)
        return interface_data

    def copy_vlans(self, data, from_mode, to_mode):
        for line in data:
            if regex.match("switchport {} allowed vlan.*".format(from_mode), line):
                self.shell.do(line.replace(from_mode, to_mode))

    def parse_interface_names(self, status_list):
        interfaces = []
        for line in status_list:
            if regex.match("(\d\S+).*", line):
                interfaces.append("ethernet {}".format(regex[0]))
            elif regex.match("ch(\d+).*", line):
                interfaces.append("port-channel {}".format(regex[0]))

        return interfaces

    def read_interface(self, interface_name):
        data = self.get_interface_data(interface_name)

        interface = Interface(name=interface_name, port_mode=ACCESS, shutdown=False)
        for line in data:
            if regex.match("switchport mode \S+", line):
                interface.port_mode = TRUNK
            if regex.match("shutdown", line):
                interface.shutdown = True
            if regex.match("switchport access vlan (\d+)", line):
                interface.access_vlan = int(regex[0])
            if regex.match("switchport general pvid (\d+)", line):
                interface.trunk_native_vlan = int(regex[0])
            if regex.match("switchport \S+ allowed vlan add (\S+)", line):
                interface.trunk_vlans += parse_vlan_ranges(regex[0])
            if regex.match("mtu (\d+)", line):
                interface.mtu = int(regex[0])

        return interface

    def reset_interface(self, interface_id):
        with self.config(), self.interface(interface_id):
            self.shell.do("no switchport access vlan")
            self.shell.do("no switchport mode")
            self.shell.do("no mtu")
        data = "\n".join(self.get_interface_data(interface_id))
        if len(data) > 0:
            raise InterfaceResetIncomplete("{}".format(data))

    def parse_interface_from_vlan_list(self, vlan_number, result):
        vlan_interfaces = []
        number = None
        for line in result:
            if regex.match('^(\d+)(.*)', line):
                number, leftover = regex
                if int(number) == vlan_number:
                    if regex.match('^\s{1,6}\S*\s+([a-zA-Z0-9-,/]+).*', leftover):
                        vlan_interfaces.extend(self.parse_interface_port_list(regex[0]))

            elif regex.match('^\s+([a-zA-Z0-9-,/]+).*', line) and int(number) == vlan_number:
                vlan_interfaces.extend(self.parse_interface_port_list(regex[0]))

        return vlan_interfaces

    def parse_interface_port_list(self, ports):
        port_list = filter(None, ports.split(','))
        interface_list = []
        for port in port_list:
            if regex.match('^ch(\d+)-(\d+)', port):
                start, end = regex
                for i in range(int(start), int(end)+1):
                    interface_list.append("port-channel {}".format(i))
            elif regex.match('^ch(\d+)', port):
                start = regex[0]
                interface_list.append("port-channel {}".format(start))
            elif regex.match('^(\d/[a-z]+)(\d+)-\d/[a-z]+(\d+)', port):
                debut, start, end = regex
                for i in range(int(start), int(end)+1):
                    interface_list.append("ethernet {0}{1}".format(debut, i))
            elif regex.match('^(\d/[a-z]+)(\d+)', port):
                debut, start = regex
                interface_list.append("ethernet {0}{1}".format(debut, start))
        return interface_list


def parse_vlan_list(result):
    vlans = []
    for line in result:
        if regex.match('^(\d+)(.*)', line):
            number, leftovers = regex
            name = None
            if regex.match('^\s{1,6}(\S+).*', leftovers):
                name = regex[0]
            vlan = Vlan(number=int(number),
                        name=name if int(number) > 1 else "default")
            vlans.append(vlan)
    return vlans


def resolve_port_mode(interface_data):
    for line in interface_data:
        if regex.match("switchport mode (\S+)", line):
            return regex[0]
        elif regex.match("switchport access vlan .*", line):
            return "access"
    return None


def assert_native_vlan_is_set(interface_id, interface_data):
    for line in interface_data:
        if regex.match("switchport general pvid (\S+)", line):
            return
    raise NativeVlanNotSet(interface_id)


def resolve_trunk_vlans(interface_data):
    for line in interface_data:
        if regex.match("switchport \S+ allowed vlan add (\S+)", line):
            return parse_vlan_ranges(regex[0])
    return []


def bond_name(number):
    return "port-channel {}".format(number)


class NamedBond(object):
    def __init__(self, number):
        self.number = number

    @property
    def name(self):
        return bond_name(self.number)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is UnknownInterface:
            raise UnknownBond(self.number), None, exc_tb



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netman.adapters.switches import cisco, juniper, dell, dell10g, brocade, arista
from netman.adapters.switches.juniper.mx import netconf as mx_netconf
from netman.adapters.switches.remote import RemoteSwitch
from netman.core.objects.flow_control_switch import FlowControlSwitch
from netman.core.objects.switch_descriptor import SwitchDescriptor

factories = {
    "arista": arista.eapi,
    "arista_http": arista.eapi_http,
    "arista_https": arista.eapi_https,
    "cisco": cisco.ssh,
    "brocade": brocade.ssh,
    "brocade_ssh": brocade.ssh,
    "brocade_telnet": brocade.telnet,
    "juniper": juniper.standard.netconf,
    "juniper_qfx_copper": juniper.qfx_copper.netconf,
    "juniper_mx": mx_netconf,
    "dell": dell.ssh,
    "dell_ssh": dell.ssh,
    "dell_telnet": dell.telnet,
    "dell10g": dell10g.ssh,
    "dell10g_ssh": dell10g.ssh,
    "dell10g_telnet": dell10g.telnet,
}


class RealSwitchFactory(object):

    def get_switch(self, hostname):
        raise NotImplemented()

    def get_anonymous_switch(self, **kwargs):
        return self.get_switch_by_descriptor(SwitchDescriptor(**kwargs))

    def get_switch_by_descriptor(self, switch_descriptor):
        if switch_descriptor.netman_server:
            return RemoteSwitch(switch_descriptor)
        return factories[switch_descriptor.model](switch_descriptor)


class FlowControlSwitchFactory(RealSwitchFactory):

    def __init__(self, switch_source, lock_factory):
        self.switch_source = switch_source
        self.lock_factory = lock_factory
        self.locks = {}

    def get_switch_by_descriptor(self, switch_descriptor):
        real_switch = super(FlowControlSwitchFactory, self).get_switch_by_descriptor(switch_descriptor)
        return FlowControlSwitch(real_switch, lock=self._get_lock(switch_descriptor))

    def _get_lock(self, switch_descriptor):
        key = switch_descriptor.hostname
        if key not in self.locks:
            self.locks[key] = self.lock_factory.new_lock(key)
        return self.locks[key]


SwitchFactory = FlowControlSwitchFactory



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netman.api.objects import base_interface, interface, Serializer, \
    Serializers
from netman.api.objects import sub_dict
from netman.core.objects.bond import Bond
from netman.core.objects.interface import Interface


__all__ = ['to_api', 'to_core']


class V2(Serializer):
    since_version = 2

    def to_api(self, bond):
        return dict(
            number=bond.number,
            link_speed=bond.link_speed,
            members=bond.members,
            **base_interface.to_api(bond)
        )

    def to_core(self, api_bond):
        params = dict(vars(base_interface.to_core(api_bond)))
        params.update(sub_dict(api_bond, 'number', 'link_speed', 'members'))
        return Bond(**params)


class V1(Serializer):
    since_version = 1

    def to_api(self, bond):
        return dict(
            number=bond.number,
            link_speed=bond.link_speed,
            members=bond.members,
            interface=interface.to_api(Interface(
                shutdown=bond.shutdown,
                port_mode=bond.port_mode,
                access_vlan=bond.access_vlan,
                trunk_native_vlan=bond.trunk_native_vlan,
                trunk_vlans=bond.trunk_vlans,
                mtu=bond.mtu
            ))
        )

    def to_core(self, api_bond):
        params = dict(vars(base_interface.to_core(api_bond['interface'])))
        params.update(sub_dict(api_bond, 'number', 'link_speed', 'members'))
        return Bond(**params)


serializers = Serializers(V1(), V2())

to_api = serializers.to_api
to_core = serializers.to_core



import unittest

from flexmock import flexmock, flexmock_teardown
from hamcrest import assert_that, is_

from netman.adapters.switches.util import PageReader


class PageReaderTest(unittest.TestCase):

    def setUp(self):
        self.reader = PageReader(
            read_while="--More-- or (q)uit",
            and_press="m",
            unless_prompt="#")

        self.shell_mock = flexmock()

    def tearDown(self):
        flexmock_teardown()

    def test_reads_without_pagination(self):
        self.shell_mock.should_receive("do").with_args("command", wait_for=("--More-- or (q)uit", "#"), include_last_line=True).once().ordered().and_return([
            "line1",
            "line2",
            "prompt#"
        ])

        result = self.reader.do(self.shell_mock, "command")

        assert_that(result, is_([
            "line1",
            "line2"
        ]))

    def test_reads_2_pages_and_removes_page_indicators_and_prompt(self):
        self.shell_mock.should_receive("do").with_args("command", wait_for=("--More-- or (q)uit", "#"), include_last_line=True).once().ordered().and_return([
            "line1",
            "line2",
            "--More-- or (q)uit"
        ])

        self.shell_mock.should_receive("send_key").with_args("m", wait_for=("--More-- or (q)uit", "#"), include_last_line=True).once().ordered().and_return([
            "line3",
            "line4",
            "--More-- or (q)uit"
        ])

        self.shell_mock.should_receive("send_key").with_args("m", wait_for=("--More-- or (q)uit", "#"), include_last_line=True).once().ordered().and_return([
            "line5",
            "line6",
            "prompt#"
        ])

        result = self.reader.do(self.shell_mock, "command")

        assert_that(result, is_([
            "line1",
            "line2",
            "line3",
            "line4",
            "line5",
            "line6"
        ]))



from hamcrest import assert_that, is_, has_item, not_
from netman.core.objects.exceptions import InterfaceNotInBond
from tests.adapters.compliance_test_case import ComplianceTestCase


class RemoveInterfaceFromBondTest(ComplianceTestCase):
    _dev_sample = "juniper"

    def setUp(self):
        super(RemoveInterfaceFromBondTest, self).setUp()
        self.interface_before = self.client.get_interface(self.test_port)
        self.client.add_bond(42)
        self.client.add_interface_to_bond(self.test_port, 42)

    def tearDown(self):
        self.janitor.remove_interface_from_bond(self.test_port)
        self.janitor.remove_bond(42)
        super(RemoveInterfaceFromBondTest, self).tearDown()

    def test_removes_interface_from_bond(self):
        self.client.remove_interface_from_bond(self.test_port)
        bond = self.client.get_bond(42)

        assert_that(bond.members, not_(has_item(self.test_port)))

    def test_removes_interface_bond_master(self):
        self.client.remove_interface_from_bond(self.test_port)
        interface = self.client.get_interface(self.test_port)

        assert_that(interface.bond_master, is_(None))

    def test_raises_if_interface_not_existent(self):
        with self.assertRaises(InterfaceNotInBond):
            self.client.remove_interface_from_bond(self.test_ports[1].name)

    def test_resets_the_interface_port_mode(self):
        self.client.remove_interface_from_bond(self.test_port)

        interface = self.client.get_interface(self.test_port)
        assert_that(self.interface_before.port_mode, is_(interface.port_mode))



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netman.core.objects import Model


class VrrpGroup(Model):
    def __init__(self, id=None, ips=None, priority=None, hello_interval=None, dead_interval=None, track_id=None,
                 track_decrement=None):
        self.id = id
        self.ips = ips or []
        self.priority = priority
        self.hello_interval = hello_interval
        self.dead_interval = dead_interval
        self.track_id = track_id
        self.track_decrement = track_decrement



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import warnings

from netman.core.objects.interface import BaseInterface, Interface


class Bond(BaseInterface):
    def __init__(self, number=None, link_speed=None, members=None, **interface):
        super(Bond, self).__init__(**interface)
        self.number = number
        self.link_speed = link_speed
        self.members = members or []

    @property
    def interface(self):
        warnings.warn('Deprecated: Use directly the members of Bond instead.', DeprecationWarning)
        return Interface(
            shutdown=self.shutdown,
            port_mode=self.port_mode,
            access_vlan=self.access_vlan,
            trunk_native_vlan=self.trunk_native_vlan,
            trunk_vlans=self.trunk_vlans,
        )



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase
from hamcrest import assert_that, is_, none
from netman.adapters.memory_session_storage import MemorySessionStorage
from netman.core.objects.exceptions import SessionAlreadyExists, UnknownSession
import mock


class SessionStorageTest(TestCase):
    def setUp(self):
        self.session_source = MemorySessionStorage()
        self.switch_descriptor = mock.Mock()

    def test_add_session(self):
        self.session_source.add('some_session', self.switch_descriptor)
        assert_that(self.session_source.get('some_session'),
                    is_(self.switch_descriptor))

    def test_get_session(self):
        self.session_source.add('some_session', self.switch_descriptor)
        assert_that(self.session_source.get('some_session'), is_(self.switch_descriptor))

    def test_get_nonexistent_session_is_none(self):
        assert_that(self.session_source.get('nonexistent_session'), is_(none()))

    def test_remove_session(self):
        self.session_source.add('some_session', self.switch_descriptor)
        self.session_source.remove('some_session')
        assert_that(self.session_source.get('some_session'), is_(none()))

    def test_add_session_that_already_exists_fails(self):
        self.session_source.add('some_session', self.switch_descriptor)
        with self.assertRaises(SessionAlreadyExists):
            self.session_source.add('some_session', self.switch_descriptor)

    def test_remove_nonexistent_session_fails(self):
        self.session_source.add('other_session', self.switch_descriptor)
        with self.assertRaises(UnknownSession):
            self.session_source.remove('some_session')



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hamcrest import assert_that, greater_than
from tests.adapters.compliance_test_case import ComplianceTestCase


class GetVersionsTest(ComplianceTestCase):
    _dev_sample = "cisco"

    def test_returns_a_dictionary_of_various_hardware_and_software_versions(self):
        versions = self.client.get_versions()

        assert_that(isinstance(versions, dict))
        assert_that(len(versions), greater_than(0))
